import db_postgres, macros, options, postgres, sequtils, strutils, times, timeutils, uuids

import ./models

import nre except toSeq
from unicode import capitalize, toLower

type
  MutateClauses = object
    columns*: seq[string]
    placeholders*: seq[string]
    values*: seq[string]

let UPPERCASE_PATTERN = re"(.)(\p{Lu})"

proc newMutateClauses(): MutateClauses =
  return MutateClauses(
    columns: @[],
    placeholders: @[],
    values: @[])

proc dbFormat(s: string): string = return s

proc dbFormat(dt: DateTime): string = return dt.formatIso8601

proc dbFormat[T](list: seq[T]): string =
  return "{" & list.mapIt("'" & dbFormat(it) & "'").join(",") & "}"

proc dbFormat[T](item: T): string = return $item

proc createParseStmts(t: NimNode, value: string): NimNode =
  result = newStmtList()

  if t.typeKind == ntyObject:
    if t.getType == UUID.getType:
      result.add quote do:
        discard parseUUID(`value`)
    elif t.getType == DateTime.getType:
      result.add quote do:
        discard `value`.parseIso8601

  elif t.typeKind == ntyString:
    result.add quote do:
      discard `value`

template walkFieldDefs(t: NimNode, body: untyped) =
  let tTypeImpl = t.getTypeImpl

  var nodeToItr: NimNode
  if tTypeImpl.typeKind == ntyObject: nodeToItr = tTypeImpl[2]
  elif tTypeImpl.typeKind == ntyTypeDesc: nodeToItr = tTypeImpl.getType[1].getType[2]
  else: error $t & " is not an object or type desc (it's a " & $tTypeImpl.typeKind & ")."

  for fieldDef {.inject.} in nodeToItr.children:
    # ignore AST nodes that are not field definitions
    if fieldDef.kind == nnkIdentDefs:
      let fieldIdent {.inject.} = fieldDef[0]
      let fieldType {.inject.} = fieldDef[1]
      body

    elif fieldDef.kind == nnkSym:
      let fieldIdent {.inject.} = fieldDef
      let fieldType {.inject.} = fieldDef.getType
      body

macro populateMutateClauses(t: typed, newRecord: bool, mc: var MutateClauses): untyped =

  result = newStmtList()

  # iterate over all the object's fields
  t.walkFieldDefs:

      # grab the field, it's string name, and it's type
      let fieldName = $fieldIdent

      # we do not update the ID, but we do check: if we're creating a new
      # record, we should not have an existing ID
      if fieldName == "id":
        result.add quote do:
          if `newRecord` and not `t`.id.isZero:
            raise newException(
              AssertionError,
              "Trying to create a new record, but the record already has an ID.")

      # if we're looking at an optional field, add logic to check for presence
      elif fieldType.kind == nnkBracketExpr and
         fieldType.len > 0 and
         fieldType[0] == Option.getType:

        result.add quote do:
          `mc`.columns.add(identNameToDb(`fieldName`))
          if `t`.`fieldIdent`.isSome:
            `mc`.placeholders.add("?")
            `mc`.values.add(dbFormat(`t`.`fieldIdent`.get))
          else:
            `mc`.placeholders.add("NULL")

      # otherwise assume we can convert and go ahead.
      else:
        result.add quote do:
          `mc`.columns.add(identNameToDb(`fieldName`))
          `mc`.placeholders.add("?")
          `mc`.values.add(dbFormat(`t`.`fieldIdent`))

  #echo result.repr

# TODO
macro rowToModel(modelType: typed): untyped =
  result = newStmtList()

  #echo modelType.getType[1].getType.treeRepr
  modelType.walkFieldDefs:
    result.add createParseStmts(fieldType, "")
  #[
  result.add quote do:
    User(
      id: genUUID
  #modelType.walkFieldDefs:
  ]#

macro modelName(model: typed): string =
  return $model.getTypeInst

proc identNameToDb(name: string): string =
  return name.replace(UPPERCASE_PATTERN, "$1_$2").toLower()

proc dbNameToIdent(name: string): string =
  let parts = name.split("_")
  return @[parts[0]].concat(parts[1..^1].mapIt(capitalize(it))).join("")

proc tableName[T](rec: T): string =
  return modelName(rec).identNameToDb & "s"

proc createRecord[T](db: DbConn, rec: T): T =
  var mc = newMutateClauses()
  populateMutateClauses(rec, true, mc)

  # Confusingly, getRow allows inserts and updates. We use it to get back the ID
  # we want from the row.
  let newIdStr = db.getValue(sql(
    "INSERT INTO " & tableName(rec) &
    " (" & mc.columns.join(",") & ") " &
    " VALUES (" & mc.placeholders.join(",") & ") " &
    " RETURNING id"), mc.values)

  result = rec
  result.id = parseUUID(newIdStr)
  
proc updateRecord[T](db: DbConn, rec: T): bool =
  var mc = newMutateClauses()
  populateMutateClauses(rec, false, mc)

  let setClause = zip(mc.columns, mc.placeholders).mapIt(it.a & " = " it.b).join(',')
  let numRowsUpdated = db.execAffectedRows(sql(
    "UPDATE " & tableName(rec) &
    " SET " & setClause &
    " WHERE id = ? "), mc.values.concat(@[rec.id]))

  return numRowsUpdated > 0;

# TODO
#proc getRecord[T](db: DbConn, UUID id): T =

macro listFields(t: typed): untyped =
  var fields: seq[tuple[n: string, t: string]] = @[]
  t.walkFieldDefs:
    if fieldDef.kind == nnkSym: fields.add((n: $fieldIdent, t: fieldType.repr))
    else: fields.add((n: $fieldIdent, t: $fieldType))

  result = newLit(fields)


# proc create: Typed create methods for specific records
proc createUser*(db: DbConn, user: User): User = db.createRecord(user)
proc createApiToken*(db: DbConn, token: ApiToken): ApiToken = db.createRecord(token)
proc createMeasure*(db: DbConn, measure: Measure): Measure = db.createRecord(measure)
proc createValue*(db: DbConn, value: Value): Value = db.createRecord(value)

when isMainModule:

  rowToModel(User)

  let u = User(
    displayName: "Bob",
    email: "bob@bobsco.com",
    hashedPwd: "test")

#[
  let db = open("", "", "", "host=localhost port=5500 dbname=personal_measure user=postgres password=password")
  for row in db.fastRows(sql"SELECT * FROM users"):
    echo $row
    echo "----"
    rowToModel(User)

  echo "New user:\n\t" & $db.createUser(u)
  for row in db.fastRows(sql"SELECT * FROM users"):
    echo $row
]#
